open Util

open Pos

open Format

open Kernel

open Big_int

type t = t' loc
and t' =
  | Empty
  | Underscore
  | Var of string
  | Const of Const.t
  | Kernel_const of Const.t
  | Tok of string * t
  | Call of string * t
  | Exists of string list * t
  | Exists_comma of string list
  | Forall of string list * t
  | Concat of t * t
  | Curly of t
  | Par of t
  | Colon of t * t
  | Semicolon of t * t
  | Comma of t * t
  | Dot of t * t
  | Dotdot of t * t
  | Dotdotdot of t
  | Do of t
  | Module of t
  | Eq of t * t
  | Neq of t * t
  | Gt of t * t
  | Ge of t * t
  | Lt of t * t
  | Le of t * t
  | Colon_eq of t * t
  | Namespace of string list
  | Let of t * t
  | If of string list * t * t
  | IfElse of string list * t * t * t
  | With of t * t
  | Arrow of t * t
  | Dbl_arrow of t * t
  | Left_arrow of t * t
  | Not of t
  | And of t * t
  | Or of t * t
  | For of t * t * t
  | Get of t * t
  | Function of t
  | Plus of t * t
  | Minus of t * t
  | Mul of t * t
  | Div of t * t
  | Mod of t * t
  | Pow of t * t
  | Uminus of t
  | Atom of string
  | Functor of string * t
  | This
  | List of t
  | Xml of t * t * t
  | Save_env
  | Load_env of t

open Kernel

let rec isolate_module a = 
  match a.v with
    | Dot (m, r) ->
        let v = Var.fresh () in
        [(v, m)], { a with v = Dot ({ m with v = Var v }, r) }
    | _ ->
        [], a

let rec enum_fv_term vars a =
  match a.v with
    | Var s -> Var.Set.add s vars
    | Eq (l, r)
    | Concat (l, r)
    | Comma (l, r) -> enum_fv_term (enum_fv_term vars l) r
    | Dot (_, a)
    | Functor (_, a)
    | List a
    | Par a -> enum_fv_term vars a
    | _ -> vars

let rec enum_fv_tok vars a =
  match a.v with
    | Dot (_, a) -> enum_fv_tok vars a
    | Tok (_, a) -> enum_fv_term vars a
    | _ -> vars

let rec list_of_concat a =
  match a.v with
    | Empty -> []
    | Concat (l, {v = Empty }) ->
        list_of_concat l
    | Concat (l, r) ->
        l :: list_of_concat r
    | _ -> [a]

let rec list_of_semicolon a =
  match a.v with
    | Empty -> []
    | Concat (l, {v = Empty }) ->
        list_of_semicolon l
    | Semicolon (l, {v = Empty }) ->
        list_of_semicolon l
    | Semicolon ({ v = Concat (l, {v = Empty })}, r) ->
        l :: list_of_semicolon r
    | Semicolon (l, r) ->
        l :: list_of_semicolon r
    | _ -> [a]

let rec list_of_comma a =
  match a.v with
    | Empty -> []
    | Concat (l, {v = Empty }) ->
        list_of_comma l
    | Comma (l, {v = Empty }) ->
        list_of_comma l
    | Comma ({ v = Concat (l, {v = Empty })}, r) ->
        l :: list_of_comma r
    | Comma (l, r) ->
        l :: list_of_comma r
    | _ -> [a]

let rec dedot a =
  match a.v with
    | Dot (q, { v = Var v }) ->
        let l = dedot q in
        l @ [ v ]
    | Var v -> [ v ]
    | _ ->
        Err.invalid_value a.range

let to_var ctx l pos =
  match l with
    | [v] ->
        begin
          try
            let r = String.Map.find v ctx in
            r := true;
            v
          with Not_found -> Files.use_module pos l
        end
    | _ -> Files.use_module pos l

let to_module ctx m l range = 
  match l with
    | [] -> m
    | l -> to_var ctx l range

let rec to_binder a =
  match a.v with
    | Concat (l, { v = Empty }) ->
        to_binder l
    | Var v -> v
    | Underscore -> Var.fresh ()
    | _ ->
        Err.should_be_a_variable a.range

let add_vars pos ctx s f =
  let add_var x (ctx, refs) =
    let used = ref false in
    let ctx = String.Map.add x used ctx in
    (ctx, (x, used) :: refs) in
  let ctx, refs = Var.Set.fold add_var s (ctx, []) in
  let r = f ctx in
  let check_unused unused (x, used) =
    if !used || x.[0] = '^' then unused
    else x :: unused in
  let unused = List.fold_left check_unused [] refs in
  if unused <> [] then
    Warn.unused_variables pos unused;
  r

let bind pos ctx a f =
  let x = to_binder a in
  add_vars pos ctx (Var.Set.singleton x) (f x)

let mk p m =
  let (vars, agent, v) = m in
  Fun.big_exists p vars (Fun.Par (agent, v))

let rec debracketize a =
  match a.v with
    | Par a0
    | Curly a0 -> debracketize a0
    | _ -> a

(*
let mk_array a =
  let l = Array.length a in
  fresh begin fun v ->
    let v = Value.Var v in
    let k = Var.fresh () and i = Var.fresh () and m = Var.fresh () in
    let agent_length =
      Fun.Ask (
        v, Compound.mk "length" [| k |],
        Fun.Tell (Value.Var k, Compound.mk "value" [| Value.Int l |])) in
    let rec mk indexing_agents i =
      if i < l then
        let agent =
          Fun.Ask (
            Value.Var m, Compound.mk (sprintf "i%d" i) [| |],
            Fun.Tell (Value.Var k, Compound.mk "value" [| a.(i) |])) in
        mk (agent :: indexing_agents) (succ i)
      else indexing_agents in
    let indexing_agents = mk [] 0 in
    let indexing_values =
      Value.Var m :: Value.Var i ::
        List.map Value.of_int (List.interval 0 (l - 1)) in
    let agent_get =
      Fun.Ask (
        v, Compound.mk "get" [| i; k |],
        Fun.Exists (
          m,
          Fun.Par (
            Fun.big_par indexing_agents,
            Fun.Call (
              Compound.mk "indexing" (Array.of_list indexing_values))))) in
    par (Fun.Par (agent_length, agent_get)) (return v)
  end
*)

let to_field_name e =
  match e.v with
    | Var f -> f
    | _ -> Err.invalid_field_name e.range

let dequal to_value ctx m a =
  match a.v with
    | Colon (l, r) ->
        let c = Tok (to_field_name r, { r with v = Empty }) in
        (to_value ctx m l, { r with v = c })
    | Dot (l, r) ->
        (to_value ctx m l, r)
    | _ ->
        (m, a)

let to_tok pos to_value ctx m a =
  match a.v with
    | Tok (f, args) ->
        let args = List.map (to_value ctx m) (list_of_comma args) in
        Compound.mk f (Array.of_list args)
    | _ ->
        Err.syntax pos

let rec sequentialize to_value ctx m result l pos =
  match l with
    | [] -> Fun.Tok (pos, true, result, Compound.mk "done" [| |])
    | { v = Left_arrow (patt, v) } :: t ->
        let vars = enum_fv_term Var.Set.empty patt in
        let patt =
          add_vars patt.range ctx vars begin fun ctx' ->
            to_value ctx' m patt
          end in
        let x = Var.fresh () in
        let vx = Fun.Expr.Var (pos, x) in
        add_vars pos ctx vars begin fun ctx' ->
          let v = to_value ctx m v in
          let k =
            Fun.If (
              pos, vars, Fun.Expr.Eq (pos, patt, vx),
              sequentialize to_value ctx' m result t pos,
              Fun.Skip) in
          Fun.Ask (pos, v, Compound.mk "value" [| x |], k)
        end
    | h :: t ->
        let v = to_value ctx m h in
        let k = sequentialize to_value ctx m result t pos in
        Fun.Ask (pos, v, Compound.mk "done" [| |], k)

let vars_set pos l =
  let add_var s v =
    if Var.Set.mem v s then
      begin
        Warn.variable_introduced_twice pos v;
        s
      end
    else
      Var.Set.add v s in
  List.fold_left add_var Var.Set.empty l

let check_closed e =
  match e.v with
    | Dotdotdot e -> false, e
    | _ -> true, e

let make_record to_value closed pos ctx m l =
  let add_field pos f v map =
    if f = "type" || f = "field_count" then
      Err.reserved_field_name pos f;
    if String.Map.mem f map then
      Err.duplicated_field pos f
    else
      String.Map.add f v map in
  let make_field map e =
    match e.v with
      | Colon (fe, value) ->
          let f = to_field_name fe in
          add_field fe.range f (to_value ctx m value) map
      | Var f ->
          add_field e.range f (Fun.Expr.Var (pos, to_var ctx [f] pos)) map
      | _ ->
          Err.invalid_field e.range in
  Fun.Expr.Record (pos, closed, List.fold_left make_field String.Map.empty l)
  
let to_record to_value pos ctx m e =
  let closed, e = check_closed e in
  let l = list_of_semicolon e in
  make_record to_value closed pos ctx m l

let to_ask to_agent to_value vars ctx m a =
  add_vars a.range ctx vars begin fun ctx ->
    let prepare_ask c a =
      let l = list_of_concat (debracketize c) in
      let g = List.map (fun g -> (g.range, to_value ctx m g)) l in
      let a = to_agent ctx m a in
      g, a in
    match a.v with
      | Dbl_arrow (c, a0) ->
          let g, a0 = prepare_ask c a0 in
          Fun.PersistentAsk (a.range, vars, g, a0)
      | Arrow (c, a0) ->
          let g, a0 = prepare_ask c a0 in
          Fun.TransientAsk (a.range, vars, g, a0)
      | _ ->
          Err.invalid_agent a.range
  end

let check_empty v =
  match v.v with
    | Empty -> ()
    | _ -> Err.should_be_empty v.range

let remove_empty_concat v =
  match v.v with
    | Concat (v, { v = Empty }) -> v
    | _ -> v

let to_xml_tag to_value ctx m tag v2 =
  match tag.v with
    | Colon
        ({ v = Var ns; range = ns_range } as ns_expr,
         { v = Var tag_name; range = tag_name_range }) ->
        begin
          match (remove_empty_concat v2).v with
            | Empty -> ()
            | Colon ({ v = Var ns' }, { v = Var tag_name' })
                when ns = ns' && tag_name = tag_name' -> ()
            | _ ->
                Err.mismatch_xml_tag_names
                  { first = tag.range.first; last = v2.range.last }
        end;
        let c = Compound.mk tag_name [| |] in
        let v = Fun.Expr.Compound (tag_name_range, c) in
        Some ns_expr, v
    | Var tag_name ->
        begin
          match (remove_empty_concat v2).v with
            | Empty -> ()
            | Var tag_name' when tag_name = tag_name' -> ()
            | _ ->
                Err.mismatch_xml_tag_names
                  { first = tag.range.first; last = v2.range.last }
        end;
        let c = Compound.mk tag_name [| |] in
        let v = Fun.Expr.Compound (tag.range, c) in
        None, v
    | Colon (ns, tag_name) ->
        check_empty v2;
        let tag_name = to_value ctx m tag_name in
        Some ns, tag_name
    | _ ->
        check_empty v2;
        None, to_value ctx m tag

let split_field f =
  match f.v with
    | Eq ({ v = Colon (ns, f) }, v) -> Some ns, f, Some v
    | Eq (f, v) -> None, f, Some v
    | Colon (ns, f) -> Some ns, f, None
    | _ -> None, f, None

let split_ns_fields (default_ns, ns_list, ctx, other_fields) (ns, f, v) =
  match ns, f.v with
    | Some { v = Var "xmlns" }, _ ->
        let f =
          match f.v with
            | Var f -> f
            | _ -> Err.should_be_a_variable f.range in
        let ns_list = (f, v) :: ns_list in
        let ctx = String.Map.add f (ref false) ctx in
        (default_ns, ns_list, ctx, other_fields)
    | None, Var "xmlns" ->
        (Some v, ns_list, ctx, other_fields)
    | _ ->
        (default_ns, ns_list, ctx, (ns, f, v) :: other_fields)

let rec filter_args fields =
  match fields with
    | [] -> []
    | (None, ({ v = Var _ } as f), Some v) :: tail ->
        let v =
          match v.v with
            | Atom s -> { v with v = Const (Const.String s) }
            | _ -> v in
        {f with v = Colon (f, v)} :: filter_args tail
    | _ -> failwith "filter_args"

let rec to_xml to_value pos ctx m v0 v1 v2 =
  let closed, v0 = check_closed v0 in
  let l = list_of_concat v0 in
  let tag, fields =
    match l with
      | tag :: fields -> tag, fields
      | _ -> Err.missing_xml_tag_name v0.range in
  let ns, tag_name = to_xml_tag to_value ctx m tag v2 in
  let fields = List.map split_field fields in
  let (default_ns, ns_list, ctx, other_fields) =
    List.fold_left split_ns_fields (None, [], ctx, []) fields in
  let l = filter_args other_fields in
  let args = make_record to_value closed pos ctx m l in
  let l = list_of_concat v1 in
  let contents =
    Fun.Expr.List (pos, List.map (mk_xml_contents to_value ctx m) l) in
  let ns =
    match ns with
      | None -> Fun.Expr.Const (pos, Const.String "")
      | Some ns -> to_value ctx m ns in
  let xml = Compound.mk "xml" [| ns; tag_name; args; contents |] in
  Fun.Expr.Compound (pos, xml)
and mk_xml_contents to_value ctx m v =
  match v.v with
    | Const (Const.String s) -> Fun.Expr.Const (v.range, Const.String s)
    | Curly v -> to_value ctx m v
    | Xml (v0, v1, v2) -> to_xml to_value v.range ctx m v0 v1 v2
    | _ -> failwith "mk_xml_contents"

let env = ref (0, [])

let env_name env_count = sprintf "env%d" env_count

let rec to_value ctx m a =
  match a.v with
    | Underscore -> Fun.Expr.Underscore a.range
    | This -> m
    | Concat (l, { v = Empty }) -> to_value ctx m l
    | Concat ({ v = Concat (v0, v1) }, v2) ->
        to_value ctx m
          { v = Concat (v0, { v = Concat (v1, v2); range = a.range});
            range = a.range }
    | Concat (v0, { v = Concat ({ v = Concat (v1, v2) }, v3)}) ->
        to_value ctx m
          { v = Concat (
              v0,
              { v = Concat (v1, { v= Concat (v2, v3); range = a.range });
                range = a.range});
            range = a.range }
    | Exists (l, v0) | Concat ({ v = Exists_comma l }, v0) ->
        add_vars a.range ctx (vars_set a.range l) begin fun ctx' ->
          Fun.Expr.big_exists a.range l (to_value ctx' m v0)
        end
    | Const c -> Fun.Expr.Const (a.range, c)
    | Kernel_const c -> Fun.Expr.Kernel_const (a.range, c)
    | Uminus { v = (Const (Const.Int i)) } ->
        Fun.Expr.of_big_int a.range (minus_big_int i)
    | Colon (_, _) | Dot (_, { v = Tok _ }) | Tok _ ->
        let vm, a = dequal to_value ctx m a in
        let c = to_tok a.range to_value ctx m a in
        Fun.Expr.Fun (a.range, vm, c)
    | Par v -> to_value ctx m v
    | Module a ->
        let v = Var.fresh () in
        let vv = Fun.Expr.Var (a.range, v) in
        Fun.Expr.Module (a.range, v, to_agent ctx vv a)
    | With (e1, e2) ->
        let e1 = to_value ctx m e1 in
        let v = Var.fresh () in
        let vv = Fun.Expr.Var (a.range, v) in
        Fun.Expr.Let (a.range, v, e1, to_value ctx vv e2)
    | Do a ->
        let a' = debracketize a in
        let l = list_of_concat a' in
        let v = Var.fresh () in
        let result = Fun.Expr.Var (a.range, v) in
        let s = sequentialize to_value ctx m result l a.range in
        Fun.Expr.Exists (a.range, v, Fun.Expr.Where (a.range, result, s))
    | Atom f ->
        let c = Compound.mk f [| |] in
        Fun.Expr.Compound (a.range, c)
    | Functor (f, args) ->
        let args = List.map (to_value ctx m) (list_of_comma args) in
        let c = Compound.mk f (Array.of_list args) in
        Fun.Expr.Compound (a.range, c)
    | Comma (l, r) ->
        let args = List.map (to_value ctx m) (l :: list_of_comma r) in
        let c = Compound.mk "" (Array.of_list args) in
        Fun.Expr.Compound (a.range, c)
    | List l ->
        Fun.Expr.List (a.range, List.map (to_value ctx m) (list_of_comma l))
    | Concat (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Concat (a.range, v0, v1)
    | Not v0 ->
        let v0 = to_value ctx m v0 in
        Fun.Expr.Not (a.range, v0)
    | And (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.And (a.range, v0, v1)
    | Or (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Or (a.range, v0, v1)
    | Plus (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "add" a.range [v0; v1]
    | Minus (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "sub" a.range [v0; v1]
    | Mul (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "mul" a.range [v0; v1]
    | Div (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "div" a.range [v0; v1]
    | Mod (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "mod" a.range [v0; v1]
    | Pow (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "pow" a.range [v0; v1]
    | Colon_eq (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Fun (a.range, v0, Compound.mk "set" [| v1 |])
    | Eq (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Eq (a.range, v0, v1)
    | Neq (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Neq (a.range, v0, v1)
    | Gt (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "gt" a.range [v0; v1]
    | Ge (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "ge" a.range [v0; v1]
    | Lt (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "lt" a.range [v0; v1]
    | Le (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "arith"] "le" a.range [v0; v1]
    | Get (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.Fun (a.range, v0, Compound.mk "get" [| v1 |])
    | Dotdot (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Expr.lib_call ["data"; "int"] "range" a.range [v0; v1]
    | For (patt, v0, v1) ->
        let vars = enum_fv_term Var.Set.empty patt in
        let patt =
          add_vars patt.range ctx vars begin fun ctx' ->
            to_value ctx' m patt
          end in
        let v0 = to_value ctx m v0 in
        let v1 =
          add_vars v1.range ctx vars begin fun ctx' ->
            to_value ctx' m v1
          end in
        let x = Var.fresh () in
        let vx = Fun.Expr.Var (a.range, x) in
        let v = Var.fresh () in
        let vv = Fun.Expr.Var (a.range, v) in
        let k = Var.fresh () in
        let vk = Fun.Expr.Var (a.range, k) in
        let a1 =
          Fun.Ask
            (a.range, vv, Compound.mk "do" [| x; k |],
             Fun.If (
               a.range, vars, Fun.Expr.Eq (a.range, patt, vx),
               Fun.Eq (a.range, vk, v1),
               Fun.Tok (a.range, true, vk, Compound.mk "done" [| |]))) in
        let v1 = Fun.Expr.Module (a.range, v, a1) in
        Fun.Expr.Fun (a.range, v0, Compound.mk "iter" [| v1 |])
    | If (vars, cond, case_true) ->
        let vars = vars_set a.range vars in
        add_vars a.range ctx vars begin fun ctx' ->
          let cond = to_value ctx' m cond in
          let case_true = to_value ctx' m case_true in
          let case_false = Fun.Expr.Underscore a.range in
          Fun.Expr.If (a.range, vars, cond, case_true, case_false)
        end
    | IfElse (vars, cond, case_true, case_false) ->
        let vars = vars_set a.range vars in
        add_vars a.range ctx vars begin fun ctx' ->
          let cond = to_value ctx' m cond in
          let case_true = to_value ctx' m case_true in
          let case_false = to_value ctx m case_false in
          Fun.Expr.If (a.range, vars, cond, case_true, case_false)
        end
    | Curly e ->
        to_record to_value a.range ctx m e
    | Xml (v0, v1, v2) ->
        Fun.Expr.List (a.range, [to_xml to_value a.range ctx m v0 v1 v2])
    | _ ->
        Fun.Expr.Var (a.range, to_var ctx (dedot a) a.range)
and to_agent ctx m a =
  match a.v with
    | Empty -> Fun.Skip
    | Exists (l, a0) | Concat ({ v = Exists_comma l }, a0) ->
        add_vars a.range ctx (vars_set a.range l) begin fun ctx' ->
          Fun.big_exists a.range l (to_agent ctx' m a0)
        end
    | Colon _ | Dot _ | Tok _ ->
        let vm, a = dequal to_value ctx m a in
        let c = to_tok a.range to_value ctx m a in
        Fun.Tok (a.range, true, vm, c)
    | Concat ({ v = Concat (v0, v1) }, v2) ->
        to_agent ctx m
          { v = Concat (v0, { v = Concat (v1, v2); range = a.range});
            range = a.range }
    | Concat (v0, { v = Concat ({ v = Concat (v1, v2) }, v3)}) ->
        to_agent ctx m
          { v = Concat (
              v0,
              { v = Concat (v1, { v= Concat (v2, v3); range = a.range });
                range = a.range});
            range = a.range }
    | Concat (
        { v = (Dot _ | Tok _) } as tok,
        { v = Concat ({ v = Curly body}, a1) }) ->
        let vars = enum_fv_tok Var.Set.empty tok in
        let l, tok = isolate_module tok in
        let ctx' =
          List.fold_left (fun set (x, _) -> String.Map.add x (ref false) set) ctx l in
        add_vars tok.range ctx' vars begin fun ctx' ->
          let g = to_value ctx' m tok in
          let body = to_agent ctx' m body in
          let a0 = Fun.PersistentAsk (a.range, vars, [(tok.range, g)], body) in
          let l = List.map (fun (x, v) -> (x, to_value ctx m v)) l in
          let a0 = Fun.big_let a.range l a0 in
          let a1 = to_agent ctx m a1 in
          Fun.Par (a0, a1)
        end
    | Concat ({ v = Let (patt, value)} as let_item, a0) ->
        let vars = enum_fv_term Var.Set.empty patt in
        let patt =
          add_vars patt.range ctx vars begin fun ctx' ->
            to_value ctx' m patt
          end in
        let (sub_agent, other) =
          match a0.v with
            | Concat ({ v = (Curly _ | Par _)} as sub_agent, other) ->
                (sub_agent, other)
            | _ ->
                (a0, { a0 with v = Empty }) in
        let value = to_value ctx m value in
        let sub_agent =
          add_vars let_item.range ctx vars begin fun ctx' ->
            to_agent ctx' m sub_agent
          end in
        let other = to_agent ctx m other in
        let x = Var.fresh () in
        let ex = Fun.Expr.Var (a.range, x) in
        Fun.Par (
          Fun.Let(
            a.range, x, value,
            Fun.If (
              a.range, vars, Fun.Expr.Eq (a.range, patt, ex),
              sub_agent, Fun.Skip)),
          other)
    | Concat (
        { v = (Function tok) } as hd,
        { v = Concat ({ v = Curly body}, a1) }) ->
        let vars = enum_fv_tok (Var.Set.singleton "result") tok in
        let (l, tok) = isolate_module tok in
        let ctx' =
          List.fold_left (fun set (x, _) -> String.Map.add x (ref false) set) ctx l in
        add_vars hd.range ctx' vars begin fun ctx' ->
          let (vm, tok) = dequal to_value ctx' m tok in
          let c = to_tok a.range to_value ctx' m tok in
          let f = Compound.f c in
          let args = c.Compound.args in
          let vresult = Fun.Expr.Var (a.range, "result") in
          let t = Compound.mk f (Array.append args [| vresult |]) in
          let body = to_agent ctx' m body in
          let g = Fun.Expr.Fun (a.range, vm, t) in
          let r = Fun.PersistentAsk (a.range, vars, [a.range, g], body) in
          let l = List.map (fun (x, v) -> (x, to_value ctx m v)) l in
          let a1 = to_agent ctx m a1 in
          Fun.Par (Fun.big_let a.range l r, a1)
        end
    | Concat (a0, a1) ->
        let a0' = to_agent ctx m a0 in
        let a1' = to_agent ctx m a1 in
        Fun.Par (a0', a1')
    | Call (f, args) ->
        let args = List.map (to_value ctx m) (list_of_comma args) in
        Fun.Call (a.range, Compound.mk f (Array.of_list args))
    | Curly a0 | Par a0 ->
        to_agent ctx m a0
    | With (e1, a2) ->
        let e1 = to_value ctx m e1 in
        let v = Var.fresh () in
        let vv = Fun.Expr.Var (a.range, v) in
        Fun.Let (a.range, v, e1, to_agent ctx vv a2)
    | Do _ ->
        Fun.Ask (
          a.range,
          to_value ctx m a,
          Compound.mk "done" [| |],
          Fun.Tok (
            a.range, false, Fun.Expr.Var (a.range, "_"),
            Compound.mk "dummy" [| |]))
    | Eq (v0, v1) ->
        let v0 = to_value ctx m v0 in
        let v1 = to_value ctx m v1 in
        Fun.Eq (a.range, v0, v1)
    | If (vars, cond, case_true) ->
        let vars = vars_set a.range vars in
        add_vars a.range ctx vars begin fun ctx' ->
          let cond = to_value ctx' m cond in
          let case_true = to_agent ctx' m case_true in
          Fun.If (a.range, vars, cond, case_true, Fun.Skip)
        end
    | IfElse (vars, cond, case_true, case_false) ->
        let vars = vars_set a.range vars in
        add_vars a.range ctx vars begin fun ctx' ->
          let cond = to_value ctx' m cond in
          let case_true = to_agent ctx' m case_true in
          let case_false = to_agent ctx m case_false in
          Fun.If (a.range, vars, cond, case_true, case_false)
        end
    | For (patt, v0, a1) ->
        let vars = enum_fv_term Var.Set.empty patt in
        let patt =
          add_vars patt.range ctx vars begin fun ctx' ->
            to_value ctx' m patt
          end in
        let v0 = to_value ctx m v0 in
        let a1 =
          add_vars a1.range ctx vars begin fun ctx' ->
            to_agent ctx' m a1
          end in
        let x = Var.fresh () in
        let vx = Fun.Expr.Var (a.range, x) in
        let v = Var.fresh () in
        let vv = Fun.Expr.Var (a.range, v) in
        let a1 =
          Fun.Ask
            (a.range, vv, Compound.mk "do" [| x |],
             Fun.If (
               a.range, vars, Fun.Expr.Eq (a.range, patt, vx), a1, Fun.Skip)) in
        let v1 = Fun.Expr.Module (a.range, v, a1) in
        Fun.Tok (a.range, true, v0, Compound.mk "iter" [| v1 |])
    | Function a0 ->
        begin
          match a0.v with
            | Eq (tok, v) ->
                let vars = enum_fv_tok Var.Set.empty tok in
                let (l, tok) = isolate_module tok in
                let ctx' =
                  List.fold_left
                    (fun set (x, _) -> String.Map.add x (ref false) set)
                    ctx l in
                add_vars tok.range ctx' vars begin fun ctx' ->
                  let (vm, tok) = dequal to_value ctx' m tok in
                  let c = to_tok a.range to_value ctx' m tok in
                  let f = Compound.f c in
                  let args = c.Compound.args in
                  let result = Var.fresh () in
                  let vresult = Fun.Expr.Var (a.range, result) in
                  let t = Compound.mk f (Array.append args [| vresult |]) in
                  let v = to_value ctx' m v in
                  let body =
                    Fun.Eq (a0.range, Fun.Expr.Var (a.range, result), v) in
                  let g = Fun.Expr.Fun (a.range, vm, t) in
                  let r = Fun.PersistentAsk (
                    a.range, Var.Set.add result vars, [a.range, g], body) in
                  let l = List.map (fun (x, v) -> (x, to_value ctx m v)) l in
                  Fun.big_let a.range l r
                end
            | _ ->
                Err.should_be_equality a0.range
        end
    | Forall (vars, { v = (Curly a0 | Par a0) }) ->
        to_ask to_agent to_value (vars_set a.range vars) ctx m a0
    | Forall (vars, { v = Concat (a0, a1) }) ->
        let a0 = to_ask to_agent to_value (vars_set a.range vars) ctx m a0 in
        let a1 = to_agent ctx m a1 in
        Fun.Par (a0, a1)
    | Load_env a0 ->
        let (env_count, env_list) = !env in
        let add ctx x = String.Map.add x (ref false) ctx in
        let ctx = List.fold_left add ctx env_list in
        let a0 = to_agent ctx m a0 in
        if env_list = [] then
          a0
        else
          begin
            let env_array = Array.of_list env_list in
            Fun.Ask (
              a.range, Fun.Expr.Var (a.range, "$"),
              Compound.mk (env_name env_count) env_array, a0)
          end
    | Save_env ->
        let (env_count, _) = !env in
        let env_count = succ env_count in
        let env_list = String.Map.fold (fun x b l -> b := true; x :: l) ctx [] in
        env := (env_count, env_list);
        if env_list = [] then
          Fun.Skip
        else
          begin
            let mk_var v = Fun.Expr.Var (a.range, v) in
            let env_array = Array.of_list (List.map mk_var env_list) in
            Fun.Tok (
              a.range, false, mk_var "$",
              Compound.mk (env_name env_count) env_array)
          end
    | _ ->
        to_ask to_agent to_value Var.Set.empty ctx m a
